#include <pspkernel.h>
#include <pspdisplay.h>
#include <stdlib.h>
#include <pspdebug.h>
#include <pspsdk.h>
#include <pspctrl.h>
#include <pspgu.h>
#include <pspthreadman.h>
#include <stdio.h>
#include <string.h>
#include <pspmath.h>
#include "blit.h"
#include <psploadcore.h>
#include <pspiofilemgr.h>
#include <malloc.h>
#include <stdio.h>

//PSP_MODULE_INFO("Appswitcher", 0x0800, 1, 1);	// VSH Mode
PSP_MODULE_INFO("Appswitcher", 0x1000, 1, 1);	// Kernel Mode
PSP_MAIN_THREAD_ATTR(0);
PSP_MAIN_THREAD_STACK_SIZE_KB(0);

char cMessage[500];

void putlog(char *cTemp)
{
	int fp = sceIoOpen("ms0:/appswitcher.log", PSP_O_WRONLY|PSP_O_CREAT, 0777);
	if(fp > 0)
	{
		sceIoLseek(fp, 0, SEEK_END);
		sceIoWrite(fp, cTemp, strlen(cTemp));
		sceIoWrite(fp, "\n", 1);
		sceIoClose(fp);
	}
}

//------------------------------------------------------------------------------
// Revert the current State
// How a current State is identified:
// if File1 is there and File2 isn't then this is the current state
// -> rename File1 to File2 and return the current line in the textfile
// return: -2 if error
//         -1 if no current state found
//         >= 0 the line of the current state
//------------------------------------------------------------------------------
int revertCurrent()
{
	char cTemp;
	char cLine[500];
	char cFile1[500];
	char cFile2[500];
	int iLine = -1;
	int iEOF = 0;
	
#ifdef DEBUG
	putlog("--- Revert current State ---");
#endif

	int fd = sceIoOpen("ms0:/seplugins/appswitcher.txt", PSP_O_RDONLY, 0777);
	if(fd <= 0)
	{
		putlog("ERROR: ms0:/seplugins/appswitcher.txt could not be opened!");
		return -2;
	}
	else
	{
		while(!iEOF)
		{
			// a new line
			iLine++;
			memset(&cFile1, 0, 500);
			int iPos = 0;
			int i;
			for(i=0; i < 500; i++)
			{
				if(!sceIoRead(fd, &cTemp, 1))
				{
					iEOF = 1;
					break;
				}
				if(cTemp == '\n')
				{
					break;
				}
				if(cTemp != 0x0A && cTemp != 0x0D)
				{
					cFile1[iPos] = cTemp;
					iPos++;
				}
			}

			if(iEOF)
			{
				if(strlen(cFile1) < 5)
				{
					break;
				}
			}
			
			memset(&cFile2, 0, 500);
			iPos = 0;
			int iGo = 0;
			for(i=0; i < 500; i++)
			{
				if(cFile1[i] == 0x0)
				{
					break;
				}
				if(!iGo)
				{
					if(cFile1[i] == 0x3B)
					{
						cFile1[i] = 0x0;
						iGo = 1;
					}
				}
				else
				{
					cFile2[iPos] = cFile1[i];
					iPos++;
				}
			}

#ifdef DEBUG
			putlog("Line read:");
			putlog(cFile1);
			putlog(cFile2);
#endif

			if((sceIoDopen(cFile1) >= 0) && (sceIoDopen(cFile2) < 0))
			{
				// we found the current state, now reverting back
#ifdef DEBUG
				putlog("Found State, reverting back.");
#endif
				int status = sceIoRename(cFile1, cFile2);
				if(status >= 0)
				{
#ifdef DEBUG
					putlog("Reverting Success.");
#endif
					sceIoClose(fd);
					return iLine;
				}
				else
				{
#ifdef DEBUG
					putlog("Reverting Error!");
#endif
					sceIoClose(fd);
					putlog("Found current State but could not rename the following:");
					putlog(cFile1);
					putlog(cFile2);
					return -2;
				}
			}
		}

		// if we come here, the file has no current state for us
		sceIoClose(fd);
		return -1;
	}
}

//------------------------------------------------------------------------------
// Get the next State
// How a "Next" State is identified:
// if File1 isn't there and File2 is then this is the next state
// -> rename File2 to File1
// return: -2 if error while renaming
//         -1 if no next state can be found (call this function again with 0)
//         >= 0 the line of the current state, succesful
//------------------------------------------------------------------------------
int getNext(int iStartLine)
{
// TODO!
	char cTemp;
	char cLine[500];
	char cFile1[500];
	char cFile2[500];
	int iLine = -1;
	int iEOF = 0;
	
#ifdef DEBUG
	putlog(" --- Finding next State ---");
#endif

	int fd = sceIoOpen("ms0:/seplugins/appswitcher.txt", PSP_O_RDONLY, 0777);
	if(fd <= 0)
	{
		putlog("ERROR: ms0:/seplugins/appswitcher.txt could not be opened!");
		return -2;
	}
	else
	{
		while(!iEOF)
		{
			// a new line
			iLine++;
			memset(&cFile1, 0, 500);
			int iPos = 0;
			int i;
			for(i=0; i < 500; i++)
			{
				if(!sceIoRead(fd, &cTemp, 1))
				{
					iEOF = 1;
					break;
				}
				if(cTemp == '\n')
				{
					break;
				}
				if(cTemp != 0x0A && cTemp != 0x0D)
				{
					cFile1[iPos] = cTemp;
					iPos++;
				}
			}

			if(iEOF)
			{
				if(strlen(cFile1) < 5)
				{
					break;
				}
			}
			
			if(iLine >= iStartLine)
			{
				memset(&cFile2, 0, 500);
				iPos = 0;
				int iGo = 0;
				for(i=0; i < 500; i++)
				{
					if(cFile1[i] == 0x0)
					{
						break;
					}
					if(!iGo)
					{
						if(cFile1[i] == 0x3B)
						{
							cFile1[i] = 0x0;
							iGo = 1;
						}
					}
					else
					{
						cFile2[iPos] = cFile1[i];
						iPos++;
					}
				}

#ifdef DEBUG
				putlog("Line read:");
				putlog(cFile1);
				putlog(cFile2);
#endif

				if(((sceIoDopen(cFile2) >= 0) && (sceIoDopen(cFile1) < 0)) || ((sceIoDopen(cFile2) >= 0) && (strstr(cFile1, "ms0:/PSP/GAME"))))
				{
					// we found the next state, now renaming...
#ifdef DEBUG
					putlog("Next State Found!");
#endif
					if(strstr(cFile1, "ms0:/PSP/GAME"))
					{
						sceIoRmdir("ms0:/PSP/GAME");
					}
					
					int status = sceIoRename(cFile2, cFile1);
					if(status >= 0)
					{
#ifdef DEBUG
						putlog("Renaming Success.");
#endif
						sprintf(cMessage, "Switched to: %s", cFile2);

						if(sceIoDopen("ms0:/PSP/GAME") < 0)
						{
							// Folder "ms0:/PSP/GAME" MUST exist, else the PSP doesn't show anything!
							sceIoMkdir("ms0:/PSP/GAME", 0777);
						}

						sceIoClose(fd);
						return iLine;
					}
					else
					{
#ifdef DEBUG
						putlog("Renaming Error!");
#endif
						putlog("Found next State but could not rename the following:");
						putlog(cFile1);
						putlog(cFile2);
						sceIoClose(fd);
						return -2;
					}
				}
			}
		}
		// if we come here, the file has no current state for us
		sceIoClose(fd);
		return -1;
	}
}

int main_thread(SceSize args, void *argp)
{
	SceCtrlData pad, oldpad;
	int iMessage[3] = {0, 0, 200};

	sceKernelDelayThread(20*100000);
	//blit_setup();

	while(1)
	{
		sceCtrlPeekBufferPositive(&pad, 1);
		if((pad.Buttons != oldpad.Buttons) && (pad.Buttons & PSP_CTRL_RTRIGGER) && (pad.Buttons & PSP_CTRL_RIGHT))
		{
			// Next State
			int iCurrent = revertCurrent();
			if(iCurrent != -2)
			{
				int iNext = getNext(iCurrent+1);
				if(iNext == -1)
				{
					iNext = getNext(0);
					if(iNext == -2)
					{
						iMessage[1] = 200;
					}
					else
					{
						iMessage[0] = 200;
					}
				}
				else
				{
					iMessage[0] = 200;
				}
			}
			else
			{
				iMessage[1] = 200;
			}
			sceIoDevctl("fatms0:", 0x0240D81E, NULL, 0, NULL, 0 );	// flush the MS cache
		}
		oldpad = pad;

		if(iMessage[0] > 0)
		{
			iMessage[0]--;
			blit_string(1, 32, &cMessage, 0xffffff, 0x000000);
			//blit_string(1, 32, "This is just a message.", 0xffffff, 0x000000);
		}
		else if(iMessage[1] > 0)
		{
			iMessage[1]--;
			blit_string(1, 32, "There has been an ERROR! Check log for details.", 0xffffff, 0x000000);
		}
		else if(iMessage[2] > 0)
		{
			iMessage[2]--;
			blit_string(1, 32, "APPSwitcher PSP 1.1 by MK2k loaded", 0xffffff, 0x000000);
		}

		sceDisplayWaitVblankStart();
		sceKernelDelayThread(1000);
	}
	sceKernelSleepThread();
	return 0;
}

int module_start(SceSize args, void *argp) {

	int thid;

	//thid = sceKernelCreateThread("appswitcher", main_thread, 6, 0x4000, 0, NULL);
	thid = sceKernelCreateThread("appswitcher", main_thread, 15, 0x4000, 0, NULL);
	if(thid >= 0) sceKernelStartThread(thid, args, argp);

	return 0;
	
	sceKernelExitDeleteThread(0);
}

int module_stop(SceSize args, void *argp)
{
	return 0;
}
